"""
  Module to conduct connected component clustering for group-based record 
  linkage.

  This clustering method considers all the pairwise links (in the similarity 
  graph G) with a similarity greater than a user defined similarity threshold, 
  and then finds the connected components in the resulting graph. A connected 
  component is a set of nodes in G that are directly or indirectly connected via 
  edges. This technique has previously been used as a baseline technique to 
  compare more sophisticated clustering approaches.
  
  To run this code, the user must provide the similarity file  
  (all-attr-no-weights.csv, all-attr-weights.csv, 
  parent-name-address-no-weights.csv, parent-name-address-weights.csv,  
  parent-name-no-weights.csv, or parent-name-weights.csv), and the similarity
  threshold (expected to be between 0.7 and 1.0 including those values).
    
  The similarity file is of the following format:
  file format 
  -----------
    <num_rec>
    <num_links>
    rec_id_1, <optional attributes about record with rec_id_1>
    rec_id_2, <optional attributes about record with rec_id_2>
    ...
    rec_id_[num_rec], <optional attributes about record with rec_id_[num_rec]>
    rec_id_a,rec_id_b,sim_val, <optional attributes about link rec_a to rec_b>
    ...
    rec_id_x,rec_id_y,sim_val, <optional attributes about link rec_x to rec_y>
  
  Optionally, the user can provide a file specifying whether or not a given
  pair of records is plausible with respect to a certain constraint (such as
  temporal constraints).
  
  The constraint plausibility file should be of the following format:
  file format 
  -----------
      Rec_ID, <rec_id_z, ..., rec_id_c, rec_id_b, rec_id_a>
      rec_id_a       1,  ...,         ,        1
      rec_id_b        ,  ...,       1
      rec_id_c       1,  ...
      ...        ...
      rec_id_z
   
  For more information regarding the file format, please
  refer to our documentation in 'https://dmm.anu.edu.au/HISTRL/'.

  Usage: 
    python connected_component_clustering.py <similarity file> 
                <similarity threshold> <optional:constraint plausibility file>
  
  Please note that this program is written for python 2.7.
"""

import csv
import sys
import time
import networkx
import system_utils

def conn_comp_method(ind_pair_sim_dict, ind_id_never_compared_set, 
                  all_record_set, min_cluster_sim, not_plaus_dict = None):
  '''
  This function executes the connected component clustering method as described 
  above.
  It takes the original similarity dictionary (ind_pair_sim_dict)
  as input, and generates a new graph containing only the edges with
  similarity >= min_cluster_sim. Then we find the connected components 
  in this new graph where each connected component is a cluster representing 
  records belonging to a single entity (or set of entities).
  
  input
  -----
    ind_pair_sim_dict - A pairwise similarity dictionary containing a sorted 
                        record pair as the key, and the corresponding similarity
                        as the value.
                        
    ind_id_never_compared_set -  Set of record IDs in the data set which do not
                                 occur in the similarity dictionary.
                                 
    all_record_set - Set containing all records from the data set.
    
    min_cluster_sim -  The similarity threshold to consider for filtering 
                       record pairs.
                       
    not_plaus_dict - A dictionary (optional parameter)  containing a record
                    ID as the key and a set of record IDs which are not 
                    plausible to occur with the key ID according to some
                    constraint.
                          
  output
  ------
    final_cluster_list - A list of clusters, each as a tuple made of the 
                         individual identifiers in the cluster, the cluster size, 
                         and a quintuplet with the minimum, average, median, 
                         maximum and the density (ratio of the number of edges 
                         in each cluster from the similarity dictionary, divided 
                         by the total number of possible edges).
                         The clusters are also printed to a file with name 
                         conn_comp_clustering_<time>.csv, in the format 
                         described in our documentation at 
                         'https://dmm.anu.edu.au/HISTRL/'.
  '''
  
  start_time = time.time()
  
  print 'Executing connected component clustering with ' + \
  'minimum similarity threshold %.2f ....' %(min_cluster_sim)
  print
  
  # Initialize an empty graph
  G = networkx.Graph()

  graph_edge_set = set() # Set to contain graph edges
  graph_node_set = set() # Set to contain graph nodes

  G.clear()  # Remove any previous vertices and edges
  
  for ind_id_pair in ind_pair_sim_dict:
            
    first_ind, second_ind = ind_id_pair
    
    # Add the nodes to graph node set
    graph_node_set.add(first_ind)
    graph_node_set.add(second_ind)
    
    if (not_plaus_dict != None):
        not_poss = not_plaus_dict.get(first_ind)
        if not_poss != None:
            if second_ind in not_poss:
                continue # If the edge is temporally not possible, ignore it
                
    similarity = ind_pair_sim_dict[ind_id_pair]
    
    if (similarity >= min_cluster_sim): # Add the edge to the graph if its 
                                    # similarity is greater than the threshold.
        graph_edge_set.add((first_ind, second_ind))
   
  G.add_edges_from(graph_edge_set)  # Build new graph
  G.add_nodes_from(graph_node_set)

  assert len(graph_edge_set) == len(G.edges())
  assert len(graph_node_set) == len(G.nodes())

  # Get list of clusters (connected components), each a set of individuals
  #
  cluster_list = list(networkx.connected_components(G))
  
  final_cluster_list = system_utils.generate_final_cluster_tuples(cluster_list,  
                                  ind_id_never_compared_set, set(),  
                                  all_record_set, ind_pair_sim_dict)

  print 'Completed algorithm execution.'
  print 'Time used: %.2f sec' % (time.time() - start_time)
  print

  return final_cluster_list
  
# ---------------------------------------------------------------------------    


# Call the evaluation_func method for execution. Please see the documentation at
# the beginning of the code to find how to execute the program.
if __name__== "__main__":
  assert len(sys.argv) == 3 or len(sys.argv) == 4, len(sys.argv)
  
  if (len(sys.argv) == 4):
    # The similarity file, threshold and plausibility file are provided.
    # First argument is assumed to be the similarity file. Second the threshold.
    all_record_set, not_in_sim_dict_set, sim_dict = \
    system_utils.load_sim_file(sys.argv[1])
    min_cluster_sim = float(sys.argv[2])
    not_plaus_dict = system_utils.load_constraints_file(sys.argv[3])
  
    final_cluster_list = conn_comp_method(sim_dict, not_in_sim_dict_set, 
                              all_record_set, min_cluster_sim, not_plaus_dict)
    
  else:
    # The plausibility file is not given
    all_record_set, not_in_sim_dict_set, sim_dict = \
    system_utils.load_sim_file(sys.argv[1])
    min_cluster_sim = float(sys.argv[2])
    
    final_cluster_list = conn_comp_method(sim_dict, not_in_sim_dict_set, 
                                          all_record_set, min_cluster_sim)
    
  # final_cluster_list may be used for viewing statistical information 
  # about predicted clusters.
  
  # Print final clusters (excluding statistics)
  # Print the clusters to a file
  system_utils.print_cluster_prediction(final_cluster_list, min_cluster_sim, \
                                        'conn_comp_clustering')
# -----------------------------------------------------------------------------
    

